@@ -109,6 +109,22 @@ extern PyTypeObject PyGreenlet_Type;
 #define GREENLET_USE_TRACING 1
 #endif
 
+#ifndef Py_SET_REFCNT
+   /* Py_REFCNT and Py_SIZE macros are converted to functions
+   https://bugs.python.org/issue39573 */
+  #define Py_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
+#endif
+
+#ifndef _Py_DEC_REFTOTAL
+  /* _Py_DEC_REFTOTAL macro has been removed from Python 3.9 by:
+    https://github.com/python/cpython/commit/49932fec62c616ec88da52642339d83ae719e924 */
+#  ifdef Py_REF_DEBUG
+#    define _Py_DEC_REFTOTAL _Py_RefTotal--
+#  else
+#    define _Py_DEC_REFTOTAL
+#  endif
+#endif
+
 /* Weak reference to the switching-to greenlet during the slp switch */
 static PyGreenlet* volatile ts_target = NULL;
 /* Strong reference to the switching from greenlet after the switch */
@@ -344,7 +360,7 @@ static int (*g_initialstub)(void*);
 #undef slp_switch
 
 #ifndef STACK_MAGIC
-#error "greenlet needs to be ported to this platform, or teached how to detect your compiler properly."
+#error "greenlet needs to be ported to this platform, or taught how to detect your compiler properly."
 #endif /* !STACK_MAGIC */
 
 #ifdef EXTERNAL_ASM
@@ -475,6 +491,9 @@ static int g_switchstack(void)
 		PyThreadState* tstate = PyThreadState_GET();
 		current->recursion_depth = tstate->recursion_depth;
 		current->top_frame = tstate->frame;
+#if GREENLET_USE_CONTEXT_VARS
+		current->context = tstate->context;
+#endif
 #ifdef GREENLET_USE_EXC_INFO
 		current->exc_info = tstate->exc_info;
 		current->exc_state = tstate->exc_state;
@@ -506,6 +525,15 @@ static int g_switchstack(void)
 		tstate->recursion_depth = target->recursion_depth;
 		tstate->frame = target->top_frame;
 		target->top_frame = NULL;
+
+#if GREENLET_USE_CONTEXT_VARS
+		tstate->context = target->context;
+		target->context = NULL;
+		/* Incrementing this value invalidates the contextvars cache,
+		   which would otherwise remain valid across switches */
+		tstate->context_ver++;
+#endif
+
 #ifdef GREENLET_USE_EXC_INFO
 		tstate->exc_state = target->exc_state;
 		tstate->exc_info = target->exc_info ? target->exc_info : &tstate->exc_state;
@@ -820,8 +848,7 @@ static int GREENLET_NOINLINE(g_initialstub)(void* mark)
 			result = NULL;
 		} else {
 			/* call g.run(*args, **kwargs) */
-			result = PyEval_CallObjectWithKeywords(
-				run, args, kwargs);
+			result = PyObject_Call(run, args, kwargs);
 			Py_DECREF(args);
 			Py_XDECREF(kwargs);
 		}
@@ -948,6 +975,9 @@ green_traverse(PyGreenlet *self, visitproc visit, void *arg)
 	   - frames are not visited: alive greenlets are not garbage collected anyway */
 	Py_VISIT((PyObject*)self->parent);
 	Py_VISIT(self->run_info);
+#if GREENLET_USE_CONTEXT_VARS
+	Py_VISIT(self->context);
+#endif
 #ifdef GREENLET_USE_EXC_INFO
 	Py_VISIT(self->exc_state.exc_type);
 	Py_VISIT(self->exc_state.exc_value);
@@ -980,6 +1010,9 @@ static int green_clear(PyGreenlet* self)
 	   so even if it switches we are relatively safe. */
 	Py_CLEAR(self->parent);
 	Py_CLEAR(self->run_info);
+#if GREENLET_USE_CONTEXT_VARS
+	Py_CLEAR(self->context);
+#endif
 #ifdef GREENLET_USE_EXC_INFO
 	Py_CLEAR(self->exc_state.exc_type);
 	Py_CLEAR(self->exc_state.exc_value);
@@ -997,6 +1030,7 @@ static int green_clear(PyGreenlet* self)
 static void green_dealloc(PyGreenlet* self)
 {
 	PyObject *error_type, *error_value, *error_traceback;
+	Py_ssize_t refcnt;
 
 #if GREENLET_USE_GC
 	PyObject_GC_UnTrack(self);
@@ -1006,7 +1040,7 @@ static void green_dealloc(PyGreenlet* self)
 		/* Hacks hacks hacks copied from instance_dealloc() */
 		/* Temporarily resurrect the greenlet. */
 		assert(Py_REFCNT(self) == 0);
-		Py_REFCNT(self) = 1;
+		Py_SET_REFCNT(self, 1);
 		/* Save the current exception, if any. */
 		PyErr_Fetch(&error_type, &error_value, &error_traceback);
 		if (kill_greenlet(self) < 0) {
@@ -1035,11 +1069,13 @@ static void green_dealloc(PyGreenlet* self)
 		 * it would cause a recursive call.
 		 */
 		assert(Py_REFCNT(self) > 0);
-		if (--Py_REFCNT(self) != 0) {
+		
+		refcnt = Py_REFCNT(self) - 1;
+		Py_SET_REFCNT(self, refcnt);
+		if (refcnt != 0) {
 			/* Resurrected! */
-			Py_ssize_t refcnt = Py_REFCNT(self);
 			_Py_NewReference((PyObject*) self);
-			Py_REFCNT(self) = refcnt;
+			Py_SET_REFCNT(self, refcnt);
 #if GREENLET_USE_GC
 			PyObject_GC_Track((PyObject *)self);
 #endif
@@ -1055,6 +1091,9 @@ static void green_dealloc(PyGreenlet* self)
 		PyObject_ClearWeakRefs((PyObject *) self);
 	Py_CLEAR(self->parent);
 	Py_CLEAR(self->run_info);
+#if GREENLET_USE_CONTEXT_VARS
+	Py_CLEAR(self->context);
+#endif
 #ifdef GREENLET_USE_EXC_INFO
 	Py_CLEAR(self->exc_state.exc_type);
 	Py_CLEAR(self->exc_state.exc_value);
@@ -1721,6 +1760,7 @@ initgreenlet(void)
 	PyModule_AddObject(m, "GreenletExit", PyExc_GreenletExit);
 	PyModule_AddObject(m, "GREENLET_USE_GC", PyBool_FromLong(GREENLET_USE_GC));
 	PyModule_AddObject(m, "GREENLET_USE_TRACING", PyBool_FromLong(GREENLET_USE_TRACING));
+	PyModule_AddObject(m, "GREENLET_USE_CONTEXT_VARS", PyBool_FromLong(GREENLET_USE_CONTEXT_VARS));
 
 	/* also publish module-level data as attributes of the greentype. */
 	for (p=copy_on_greentype; *p; p++) {
